@@ -16,6 +16,7 @@ RTXProcessor::RTXProcessor()
 {
   furtherInput = true;
   inword = false;
+  inwblank = false;
   printingSteps = false;
   printingRules = false;
   printingBranches = false;
@@ -69,7 +70,7 @@ RTXProcessor::read(string const &filename)
   Transducer* t = new Transducer();
   t->read(in, alphabet.size()); 
 
-  map<int, int> finals;
+  multimap<int, pair<int, double>> finals;
 
   map<int, double> finalWeights = t->getFinals();
 
@@ -77,7 +78,9 @@ RTXProcessor::read(string const &filename)
   for(int i = 0, limit = Compression::multibyte_read(in); i != limit; i++)
   {
     int key = Compression::multibyte_read(in);
-    finals[key] = Compression::multibyte_read(in);
+    int rl = Compression::multibyte_read(in);
+    double wgt = Compression::long_multibyte_read(in);
+    finals.insert(make_pair(key, make_pair(rl, wgt)));
   }
 
   mx = new MatchExe2(*t, &alphabet, finals, pat_size);
@@ -281,6 +284,7 @@ RTXProcessor::popChunk()
   else
   {
     wcerr << "tried to pop Chunk but mode is " << theStack[stackIdx].mode << endl;
+    wcerr << "The most common reason for getting this error is a macro that is missing an else clause." << endl;
     exit(1);
   }
 }
@@ -293,15 +297,19 @@ RTXProcessor::stackCopy(int src, int dest)
   {
     case 0:
       theStack[dest].b = theStack[src].b;
+      theWblankStack[dest] = theWblankStack[src];
       break;
     case 1:
       theStack[dest].i = theStack[src].i;
+      theWblankStack[dest] = theWblankStack[src];
       break;
     case 2:
       theStack[dest].s = theStack[src].s;
+      theWblankStack[dest] = theWblankStack[src];
       break;
     case 3:
       theStack[dest].c = theStack[src].c;
+      theWblankStack[dest] = theWblankStack[src];
       break;
     default:
       wcerr << "Unknown StackElement mode " << theStack[src].mode;
@@ -309,10 +317,17 @@ RTXProcessor::stackCopy(int src, int dest)
   }
 }
 
+bool
+RTXProcessor::gettingLemmaFromWord(wstring attr)
+{
+    return (attr.compare(L"lem") == 0 || attr.compare(L"lemh") == 0 || attr.compare(L"whole") == 0);
+}
+
 bool
 RTXProcessor::applyRule(const wstring& rule)
 {
   stackIdx = 0;
+  vector<bool> editted = vector<bool>(currentInput.size(), false);
   const wchar_t* rule_data = rule.data();
   for(unsigned int i = 0, rule_size = rule.size(); i < rule_size; i++)
   {
@@ -572,7 +587,10 @@ RTXProcessor::applyRule(const wstring& rule)
       {
         wstring var = popString();
         wstring val = popString();
-        variables[var] = val;
+        currentBranch->stringVars[var] = val;
+        currentBranch->wblankVars[var] = theWblankStack[stackIdx+1];
+        theWblankStack[stackIdx+1].clear();
+        if(printingSteps) { wcerr << " -> " << var << " = '" << val << "'" << endl; }
       }
         break;
       case OUTPUT:
@@ -580,7 +598,6 @@ RTXProcessor::applyRule(const wstring& rule)
       {
         Chunk* ch = popChunk();
         if(ch == NULL) break; // FETCHCHUNK
-        ch->source.clear();
         if(isLinear && ch->contents.size() == 0)
         {
           bool word = true;
@@ -596,6 +613,8 @@ RTXProcessor::applyRule(const wstring& rule)
               Chunk* temp = chunkPool.next();
               temp->isBlank = false;
               temp->target = ch->target.substr(last, c-last);
+              temp->wblank = out_wblank;
+              out_wblank.clear();
               if(chunk) currentOutput.back()->contents.push_back(temp);
               else currentOutput.push_back(temp);
               last = c+1;
@@ -630,7 +649,9 @@ RTXProcessor::applyRule(const wstring& rule)
         }
         else
         {
+          ch->wblank = out_wblank;
           currentOutput.push_back(ch);
+          out_wblank.clear();
         }
       }
         break;
@@ -676,7 +697,17 @@ RTXProcessor::applyRule(const wstring& rule)
         popString(part);
         Chunk* ch = popChunk();
         if(ch == NULL) pushStack(L"");
-        else pushStack(ch->chunkPart(attr_items[part], SourceClip));
+        else
+        {
+          if(gettingLemmaFromWord(part))
+          {
+            pushStack(ch->chunkPart(attr_items[part], SourceClip), ch->wblank);
+          }
+          else
+          {
+            pushStack(ch->chunkPart(attr_items[part], SourceClip));
+          }
+        }
         if(printingSteps) { wcerr << " -> " << theStack[stackIdx].s << endl; }
       }
         break;
@@ -687,7 +718,17 @@ RTXProcessor::applyRule(const wstring& rule)
         popString(part);
         Chunk* ch = popChunk();
         if(ch == NULL) pushStack(L"");
-        else pushStack(ch->chunkPart(attr_items[part], TargetClip));
+        else
+        {
+          if(gettingLemmaFromWord(part))
+          {
+            pushStack(ch->chunkPart(attr_items[part], TargetClip), ch->wblank);
+          }
+          else
+          {
+             pushStack(ch->chunkPart(attr_items[part], TargetClip));
+          }
+        }
         if(printingSteps) { wcerr << " -> " << theStack[stackIdx].s << endl; }
       }
         break;
@@ -709,6 +750,11 @@ RTXProcessor::applyRule(const wstring& rule)
         wstring part = popString();
         if(pos >= 0)
         {
+          if(!editted[pos])
+          {
+            currentInput[pos] = currentInput[pos]->copy();
+            editted[pos] = true;
+          }
           currentInput[pos]->setChunkPart(attr_items[part], popString());
           if(printingSteps) { wcerr << " -> " << currentInput[pos]->target << endl; }
         }
@@ -720,17 +766,23 @@ RTXProcessor::applyRule(const wstring& rule)
         break;
       case FETCHVAR:
         if(printingSteps) { wcerr << "fetchvar" << endl; }
-        pushStack(variables[popString()]);
+        {
+          wstring name = popString();
+          wstring val = currentBranch->stringVars[name];
+          wstring wblank_val = currentBranch->wblankVars[name];
+          pushStack(val, wblank_val);
+          if(printingSteps) { wcerr << " -> " << name << " = " << val << endl; }
+        }
         break;
       case FETCHCHUNK:
         if(printingSteps) { wcerr << "fetchchunk" << endl; }
-        pushStack(currentBranch->variables[popInt()]);
+        pushStack(currentBranch->chunkVars[popInt()]);
         break;
       case SETCHUNK:
         if(printingSteps) { wcerr << "setchunk" << endl; }
         {
           int pos = popInt();
-          currentBranch->variables[pos] = popChunk();
+          currentBranch->chunkVars[pos] = popChunk();
         }
         break;
       case GETCASE:
@@ -781,6 +833,8 @@ RTXProcessor::applyRule(const wstring& rule)
           Chunk* ch = chunkPool.next();
           ch->isBlank = false;
           ch->target = kid->target.substr(1, j-1);
+          ch->wblank = out_wblank;
+          out_wblank.clear();
           theStack[stackIdx].c->contents.push_back(ch);
           ch = chunkPool.next();
           ch->isBlank = true;
@@ -789,6 +843,8 @@ RTXProcessor::applyRule(const wstring& rule)
         }
         else
         {
+          kid->wblank = out_wblank;
+          out_wblank.clear();
           theStack[stackIdx].c->contents.push_back(kid);
         }
         if(printingSteps) { wcerr << " -> child with surface '" << kid->target << L"' appended" << endl; }
@@ -811,10 +867,66 @@ RTXProcessor::applyRule(const wstring& rule)
         if(theStack[stackIdx+1].mode == 2)
         {
           theStack[stackIdx].c->target += theStack[stackIdx+1].s;
+          out_wblank = combineWblanks(out_wblank, theWblankStack[stackIdx+1]);
+          theWblankStack[stackIdx+1].clear();
         }
         else
         {
           theStack[stackIdx].c->target += theStack[stackIdx+1].c->target;
+          theStack[stackIdx].c->wblank += theStack[stackIdx+1].c->wblank;
+        }
+        if(printingSteps) { wcerr << " -> " << theStack[stackIdx+1].s << endl; }
+      }
+        break;
+      case APPENDSURFACESL:
+        if(printingSteps) { wcerr << "appendsurfacesl" << endl; }
+      {
+        if(theStack[stackIdx].mode != 2 && theStack[stackIdx].mode != 3)
+        {
+          wcerr << L"Cannot append non-string to chunk surface." << endl;
+          exit(EXIT_FAILURE);
+        }
+        stackIdx--;
+        if(theStack[stackIdx].mode != 3)
+        {
+          wcerr << L"Cannot APPENDSURFACESL to non-chunk." << endl;
+          exit(EXIT_FAILURE);
+        }
+        if(theStack[stackIdx+1].mode == 2)
+        {
+          theStack[stackIdx].c->source += theStack[stackIdx+1].s;
+          out_wblank = combineWblanks(out_wblank, theWblankStack[stackIdx+1]);
+          theWblankStack[stackIdx+1].clear();
+        }
+        else
+        {
+          theStack[stackIdx].c->source += theStack[stackIdx+1].c->source;
+          theStack[stackIdx].c->wblank += theStack[stackIdx+1].c->wblank;
+        }
+        if(printingSteps) { wcerr << " -> " << theStack[stackIdx+1].s << endl; }
+      }
+        break;
+      case APPENDSURFACEREF:
+        if(printingSteps) { wcerr << "appendsurfaceref" << endl; }
+      {
+        if(theStack[stackIdx].mode != 2 && theStack[stackIdx].mode != 3)
+        {
+          wcerr << L"Cannot append non-string to chunk surface." << endl;
+          exit(EXIT_FAILURE);
+        }
+        stackIdx--;
+        if(theStack[stackIdx].mode != 3)
+        {
+          wcerr << L"Cannot APPENDSURFACEREF to non-chunk." << endl;
+          exit(EXIT_FAILURE);
+        }
+        if(theStack[stackIdx+1].mode == 2)
+        {
+          theStack[stackIdx].c->coref += theStack[stackIdx+1].s;
+        }
+        else
+        {
+          theStack[stackIdx].c->coref += theStack[stackIdx+1].c->coref;
         }
         if(printingSteps) { wcerr << " -> " << theStack[stackIdx+1].s << endl; }
       }
@@ -831,7 +943,10 @@ RTXProcessor::applyRule(const wstring& rule)
         break;
       case APPENDALLINPUT:
         if(printingSteps) { wcerr << "appendallinput" << endl; }
-        theStack[stackIdx].c->contents.swap(currentInput);
+      {
+        vector<Chunk*>& vec = theStack[stackIdx].c->contents;
+        vec.insert(vec.end(), currentInput.begin(), currentInput.end());
+      }
         break;
       case BLANK:
         if(printingSteps) { wcerr << "blank" << endl; }
@@ -850,6 +965,16 @@ RTXProcessor::applyRule(const wstring& rule)
         }
       }
         break;
+      case CONJOIN:
+        if(printingSteps) { wcerr << "conjoin" << endl; }
+      {
+        Chunk* join = chunkPool.next();
+        join->isBlank = true;
+        join->isJoiner = true;
+        join->target = L"+";
+        pushStack(join);
+      }
+        break;
       case REJECTRULE:
         if(printingSteps) { wcerr << "rejectrule" << endl; }
         return false;
@@ -883,6 +1008,13 @@ RTXProcessor::applyRule(const wstring& rule)
         int rl = popInt();
         if(pos == -2)
         {
+          if(stackIdx == 0 || theStack[stackIdx].mode != 3)
+          {
+            wcerr << "Empty stack or top item is not chunk." << endl;
+            wcerr << "Check for conditionals that might not generate output" << endl;
+            wcerr << "and ensure that lists of attributes are complete." << endl;
+            exit(1);
+          }
           theStack[stackIdx].c->rule = rl;
         }
         else
@@ -908,6 +1040,7 @@ RTXProcessor::readToken(FILE *in)
 {
   int pos = 0;
   wstring cur;
+  wstring wbl;
   wstring src;
   wstring dest;
   wstring coref;
@@ -929,10 +1062,37 @@ RTXProcessor::readToken(FILE *in)
       cur += L'\\';
       cur += wchar_t(fgetwc_unlocked(in));
     }
-    else if(val == L'[')
+    else if(val == L'[' && !inword)
     {
-      cur += L'[';
-      inSquare = true;
+      val = fgetwc_unlocked(in);
+      
+      if(val == L'[')
+      {
+        inwblank = true;
+        Chunk* ret = chunkPool.next();
+        ret->target = cur;
+        ret->isBlank = true;
+        return ret;
+      }
+      else
+      {
+        cur += L'[';
+        inSquare = true;
+        
+        if(val == L'\\')
+        {
+          cur += L'\\';
+          cur += static_cast<wchar_t>(fgetwc_unlocked(in));
+        }
+        else
+        {
+          cur += val;
+          if(val == L']')
+          {
+            inSquare = false;
+          }
+        }
+      }
     }
     else if(inSquare)
     {
@@ -942,6 +1102,51 @@ RTXProcessor::readToken(FILE *in)
         inSquare = false;
       }
     }
+    else if(inwblank)
+    {
+      if(val == L']')
+      {
+        cur += val;
+        val = fgetwc_unlocked(in);
+        
+        if(val == L'\\')
+        {
+          cur += L'\\';
+          cur += static_cast<wchar_t>(fgetwc_unlocked(in));
+        }
+        else if(val == L']')
+        {
+          cur += val;
+          val = fgetwc_unlocked(in);
+          
+          if(val == L'\\')
+          {
+            cur += L'\\';
+            cur += static_cast<wchar_t>(fgetwc_unlocked(in));
+          }
+          else if(val == L'^')
+          {
+            inwblank = false;
+            cur = L"[[" + cur;
+            wbl.swap(cur);
+            inword = true;
+          }
+          else
+          {
+            wcerr << L"Parse Error: Wordbound blank should be immediately followed by a Lexical Unit -> [[..]]^..$" << endl;
+            exit(EXIT_FAILURE);
+          }
+        }
+        else
+        {
+          cur += val;
+        }
+      }
+      else
+      {
+        cur += val;
+      }
+    }
     else if(inword && (val == L'$' || val == L'/'))
     {
       if(pos == 0)
@@ -965,10 +1170,20 @@ RTXProcessor::readToken(FILE *in)
       {
         inword = false;
         Chunk* ret = chunkPool.next();
+        ret->wblank = wbl;
         ret->source = src;
         ret->target = dest;
         ret->coref = coref;
         ret->isBlank = false;
+        if(src.size() > 0 && src[0] == L'*' && dest.size() > 0 && dest[0] == L'*')
+        {
+          Chunk* ret2 = chunkPool.next();
+          ret2->target = ret->target.substr(1) + L"<UNKNOWN:INTERNAL>";
+          ret2->contents.push_back(ret);
+          ret2->rule = -1;
+          ret2->isBlank = false;
+          return ret2;
+        }
         return ret;
       }
     }
@@ -1126,7 +1341,9 @@ RTXProcessor::checkForReduce(vector<ParseNode*>& result, ParseNode* node)
         first = back->lastWord+1;
         cur->init(back, currentOutput[0], weight);
       }
-      cur->variables = node->variables;
+      cur->stringVars = node->stringVars;
+      cur->wblankVars = node->wblankVars;
+      cur->chunkVars = node->chunkVars;
       cur->id = node->id;
       if(temp.size() == 0)
       {
@@ -1144,7 +1361,9 @@ RTXProcessor::checkForReduce(vector<ParseNode*>& result, ParseNode* node)
         {
           cur = parsePool.next();
           cur->init(*it, temp.back());
-          cur->variables = (*it)->variables;
+          cur->stringVars = (*it)->stringVars;
+          cur->wblankVars = (*it)->wblankVars;
+          cur->chunkVars = (*it)->chunkVars;
           cur->firstWord = first;
           cur->lastWord = last;
           checkForReduce(res2, cur);
@@ -1189,6 +1408,8 @@ void
 RTXProcessor::outputAll(FILE* out)
 {
   unsigned int queueSize = outputQueue.size() - 1;
+  bool conjoining = false;
+  Chunk* tojoin = NULL;
   while(outputQueue.size() > 0)
   {
     Chunk* ch = outputQueue.front();
@@ -1217,7 +1438,37 @@ RTXProcessor::outputAll(FILE* out)
         ch->writeTree(treePrintMode, NULL);
         wcerr << endl;
       }
-      ch->output(out);
+      if(ch->contents.size() > 0)
+      {
+        vector<wstring> tags = ch->getTags(vector<wstring>());
+        for(auto it = ch->contents.rbegin(); it != ch->contents.rend(); it++)
+        {
+          (*it)->updateTags(tags);
+          outputQueue.push_front(*it);
+        }
+      }
+      else if(conjoining && !ch->isBlank)
+      {
+        tojoin->conjoin(ch);
+      }
+      else if(ch->isJoiner)
+      {
+        if(tojoin != NULL) conjoining = true;
+      }
+      else
+      {
+        conjoining = false;
+        if(tojoin != NULL)
+        {
+          tojoin->output(out);
+          tojoin = NULL;
+        }
+        if(ch->isBlank)
+        {
+          writeBlank(out);
+        }
+        else tojoin = ch;
+      }
     }
     else
     {
@@ -1272,6 +1523,32 @@ RTXProcessor::outputAll(FILE* out)
       }
     }
   }
+  if(tojoin != NULL) tojoin->output(out);
+  while(!blankQueue.empty())
+  {
+    if(blankQueue.front() == L" ")
+    {
+      blankQueue.pop_front();
+    }
+    else
+    {
+      writeBlank(out);
+    }
+  }
+}
+
+void
+RTXProcessor::writeBlank(FILE* out)
+{
+  if(blankQueue.empty())
+  {
+    blankQueue.push_back(L" ");
+  }
+  Chunk* blank = chunkPool.next();
+  blank->target = blankQueue.front();
+  blankQueue.pop_front();
+  blank->isBlank = true;
+  blank->output(out);
 }
 
 bool
@@ -1420,6 +1697,18 @@ RTXProcessor::filterParseGraph()
   }
   if(printingAll && treePrintMode == TreeModeLatex) wcerr << "\\end{itemize}" << endl << endl;
   if(count == N) return shouldOutput;
+  if(count > 100 && filter.size() > 0)
+  {
+    int maxbin = 100 / filter.size();
+    for(auto& it : filter)
+    {
+      for(unsigned int i = maxbin; i < it.second.size(); i++)
+      {
+        state[it.second[i]] = 0;
+        count--;
+      }
+    }
+  }
   vector<ParseNode*> temp;
   temp.reserve(count);
   for(int i = 0; i < N; i++)
@@ -1462,7 +1751,19 @@ RTXProcessor::processGLR(FILE *in, FILE *out)
   while(true)
   {
     Chunk* next = inputBuffer.front();
-    if(next->isBlank) printingAll = false;
+    if(next->isBlank)
+    {
+      printingAll = false;
+      if(!parseGraph.empty() && !next->target.empty())
+      {
+        // if parseGraph is empty, we're about to write it immediately anyway,
+        // so don't bother with the queue
+        // if target is empty, we don't want to put it in the wrong place
+        // if a rule asks for a space
+        // -D.S. 2020-09-21
+        blankQueue.push_back(next->target);
+      }
+    }
     if(printingAll)
     {
       wcerr << endl;
@@ -1474,22 +1775,43 @@ RTXProcessor::processGLR(FILE *in, FILE *out)
     inputBuffer.pop_front();
     if(parseGraph.size() == 0)
     {
+      // skip parseGraph stuff if a blank is the only thing being processed
+      if(next->isBlank)
+      {
+        next->output(out);
+        if(furtherInput)
+        {
+          inputBuffer.push_back(readToken(in));
+        }
+        if(inputBuffer.empty())
+        {
+          wcerr.flush();
+          fflush(out);
+          break;
+        }
+        continue;
+      }
       ParseNode* temp = parsePool.next();
       temp->init(mx, next);
       temp->id = ++newBranchId;
-      temp->variables = vector<Chunk*>(varCount, NULL);
+      temp->stringVars = variables;
+      temp->wblankVars = wblank_variables;
+      temp->chunkVars = vector<Chunk*>(varCount, NULL);
       checkForReduce(parseGraph, temp);
     }
     else
     {
-      mx->prepareChunk(next->source);
+      mx->prepareChunk(next->source.size() > 0 ? next->source : next->target);
+      // conditional deals with unknowns
       vector<ParseNode*> temp;
       for(unsigned int i = 0, limit = parseGraph.size(); i < limit; i++)
       {
         ParseNode* tempNode = parsePool.next();
         tempNode->init(parseGraph[i], next, true);
         tempNode->id = parseGraph[i]->id;
-        tempNode->variables = parseGraph[i]->variables;
+        tempNode->stringVars = parseGraph[i]->stringVars;
+        tempNode->wblankVars = parseGraph[i]->wblankVars;
+        tempNode->chunkVars = parseGraph[i]->chunkVars;
         checkForReduce(temp, tempNode);
       }
       parseGraph.swap(temp);
@@ -1543,21 +1865,37 @@ RTXProcessor::processGLR(FILE *in, FILE *out)
       parseGraph[0]->getChunks(outputQueue, parseGraph[0]->length-1);
       parseGraph.clear();
       outputAll(out);
+      variables = currentBranch->stringVars;
+      wblank_variables = currentBranch->wblankVars;
       fflush(out);
+      vector<wstring> wblanks;
       vector<wstring> sources;
       vector<wstring> targets;
       vector<wstring> corefs;
       vector<bool> blanks;
+      vector<bool> unknowns;
       int N = inputBuffer.size();
       for(int i = 0; i < N; i++)
       {
         Chunk* temp = inputBuffer.front();
+        if(temp->contents.size() > 0)
+        {
+          unknowns.push_back(true);
+          temp = temp->contents[0];
+        }
+        else
+        {
+          unknowns.push_back(false);
+        }
+        wblanks.push_back(temp->wblank);
         sources.push_back(temp->source);
         targets.push_back(temp->target);
         corefs.push_back(temp->coref);
         blanks.push_back(temp->isBlank);
         inputBuffer.pop_front();
       }
+      //wcerr << "clearing chunkPool, size was " << chunkPool.size() << endl;
+      //wcerr << "clearing parsePool, size was " << parsePool.size() << endl;
       chunkPool.reset();
       parsePool.reset();
       newBranchId = 0;
@@ -1569,10 +1907,18 @@ RTXProcessor::processGLR(FILE *in, FILE *out)
       for(int i = 0; i < N; i++)
       {
         Chunk* c = chunkPool.next();
+        c->wblank = wblanks[i];
         c->source = sources[i];
         c->target = targets[i];
         c->coref = corefs[i];
         c->isBlank = blanks[i];
+        if(unknowns[i])
+        {
+          Chunk* c2 = chunkPool.next();
+          c2->target = targets[i].substr(1) + L"<UNKNOWN:INTERNAL>";
+          c2->contents.push_back(c);
+          c = c2;
+        }
         inputBuffer.push_back(c);
       }
     }
@@ -1582,6 +1928,8 @@ RTXProcessor::processGLR(FILE *in, FILE *out)
       // if stream is empty, the last token is definitely a blank
       wcerr.flush();
       inputBuffer.front()->output(out);
+      blankQueue.clear();
+      inputBuffer.pop_front();
       fflush(out);
       break;
     }
@@ -1786,14 +2134,16 @@ RTXProcessor::process(FILE* in, FILE* out)
       {
         processGLR(in, out);
       }
-      fputwc_unlocked(L'\0', out);
-      int code = fflush(out);
-      if(code != 0)
-      {
-        wcerr << L"Could not flush output " << errno << endl;
-      }
+      fputc_unlocked('\0', out);
+      fflush(out);
       chunkPool.reset();
       parsePool.reset();
+      inputBuffer.clear();
+      // I'm not sure how the leading blank after a null gets into inputBuffer,
+      // but it does and clearing the buffer seems to fix the problem
+      // (in theory, clearing the buffer here should have no effect at all
+      // because processGLR() should consume everything in it)
+      // - D.S. Aug 26 2020
     }
   }
   else if(isLinear)
